@@ -41,8 +41,6 @@ G_BEGIN_DECLS
 
 #include "gdkprivate-win32.h"
 
-#undef GDK_ROOT_PARENT /* internal access is direct */
-#define GDK_ROOT_PARENT()             ((GdkWindow *) _gdk_parent_root)
 #define GDK_WINDOW_HWND(win)          (GDK_DRAWABLE_IMPL_WIN32(((GdkWindowObject *)win)->impl)->handle)
 #define GDK_PIXMAP_HBITMAP(pixmap)    (GDK_DRAWABLE_IMPL_WIN32(((GdkPixmapObject *)pixmap)->impl)->handle)
 #define GDK_DRAWABLE_IMPL_WIN32_HANDLE(d) (((GdkDrawableImplWin32 *) d)->handle)
@@ -117,11 +115,6 @@ GdkDrawable  *gdk_win32_begin_direct_draw_libgtk_only (GdkDrawable *drawable,
 						       gint *y_offset_out);
 void          gdk_win32_end_direct_draw_libgtk_only (gpointer priv_data);
 
-GdkWindow *   gdk_win32_window_foreign_new_for_display (GdkDisplay *display,
-                                                        GdkNativeWindow anid);
-GdkWindow *   gdk_win32_window_lookup_for_display (GdkDisplay *display,
-                                                   GdkNativeWindow anid);
-
 
 G_END_DECLS
 
