@@ -34,6 +34,7 @@
 
 #include "gtkapplication.h"
 #include "gtkmain.h"
+#include "gtkmarshalers.h"
 #include "gtkintl.h"
 #include "gtkprivate.h"
 
@@ -76,6 +77,8 @@ enum
 enum
 {
   ACTIVATED,
+  QUIT,
+  ACTION,
 
   LAST_SIGNAL
 };
@@ -92,14 +95,32 @@ struct _GtkApplicationPrivate
 
 G_DEFINE_TYPE (GtkApplication, gtk_application, G_TYPE_APPLICATION)
 
+static void
+process_timestamp_from_platform_data (GVariant *platform_data)
+{
+  /* TODO - extract timestamp from here, update GDK time */
+}
+
 static gboolean
-gtk_application_default_quit (GApplication *application,
-                              guint         timestamp)
+gtk_application_default_quit (GtkApplication *application)
 {
   gtk_main_quit ();
   return TRUE;
 }
 
+static gboolean
+gtk_application_default_quit_with_data (GApplication *application,
+					GVariant     *platform_data)
+{
+  gboolean result;
+
+  process_timestamp_from_platform_data (platform_data);
+  
+  g_signal_emit (application, gtk_application_signals[QUIT], 0, &result);
+
+  return result;
+}
+
 static void
 gtk_application_default_run (GApplication *application)
 {
@@ -128,7 +149,7 @@ gtk_application_default_prepare_activation (GApplication *application,
 }
 
 static void
-gtk_application_default_activated (GApplication *application,
+gtk_application_default_activated (GtkApplication *application,
                                    GVariant     *arguments)
 {
   GtkApplication *app = GTK_APPLICATION (application);
@@ -139,21 +160,34 @@ gtk_application_default_activated (GApplication *application,
 }
 
 static void
-gtk_application_default_action (GApplication *application,
-                                const gchar  *action_name,
-                                guint         timestamp)
+gtk_application_default_action (GtkApplication *application,
+				const gchar    *action_name)
 {
-  GtkApplication *app = GTK_APPLICATION (application);
   GtkAction *action;
 
-  action = gtk_action_group_get_action (app->priv->main_actions, action_name);
+  action = gtk_action_group_get_action (application->priv->main_actions, action_name);
   if (action)
     {
-      /* TODO set timestamp */
       gtk_action_activate (action);
     }
 }
 
+static GtkWindow *
+gtk_application_default_create_window (GtkApplication *application)
+{
+  return GTK_WINDOW (gtk_window_new (GTK_WINDOW_TOPLEVEL));
+}
+
+static void
+gtk_application_default_action_with_data (GApplication *application,
+					  const gchar  *action_name,
+					  GVariant     *platform_data)
+{
+  process_timestamp_from_platform_data (platform_data);
+
+  g_signal_emit (application, gtk_application_signals[ACTION], g_quark_from_string (action_name));
+}
+
 static GVariant *
 gtk_application_format_activation_data (void)
 {
@@ -176,11 +210,32 @@ gtk_application_format_activation_data (void)
   return g_variant_builder_end (&builder);
 }
 
+static GVariant *
+variant_from_argv (int    argc,
+		   char **argv)
+{
+  GVariantBuilder builder;
+  int i;
+
+  g_variant_builder_init (&builder, G_VARIANT_TYPE ("aay"));
+
+  for (i = 1; i < argc; i++)
+    {
+      guint8 *argv_bytes;
+
+      argv_bytes = (guint8*) argv[i];
+      g_variant_builder_add_value (&builder,
+				   g_variant_new_byte_array (argv_bytes, -1));
+    }
+  
+  return g_variant_builder_end (&builder);
+}
+
 /**
  * gtk_application_new:
+ * @appid: System-dependent application identifier
  * @argc: (allow-none) (inout): System argument count
  * @argv: (allow-none) (inout): System argument vector
- * @appid: System-dependent application identifier
  *
  * Create a new #GtkApplication, or if one has already been initialized
  * in this process, return the existing instance. This function will as
@@ -194,14 +249,15 @@ gtk_application_format_activation_data (void)
  * Since: 3.0
  */
 GtkApplication*
-gtk_application_new (gint          *argc,
-                     gchar       ***argv,
-                     const gchar   *appid)
+gtk_application_new (const gchar   *appid,
+		     gint          *argc,
+                     gchar       ***argv)
 {
   GtkApplication *app;
   gint argc_for_app;
   gchar **argv_for_app;
-  GVariant *platform_data;
+  GVariant *argv_variant;
+  GError *error = NULL;
 
   gtk_init (argc, argv);
 
@@ -214,12 +270,20 @@ gtk_application_new (gint          *argc,
   else
     argv_for_app = NULL;
 
-  app = g_object_new (GTK_TYPE_APPLICATION, "application-id", appid, NULL);
+  argv_variant = variant_from_argv (argc_for_app, argv_for_app);
 
-  platform_data = gtk_application_format_activation_data ();
-  g_application_register_with_data (G_APPLICATION (app), argc_for_app, argv_for_app,
-				    platform_data);
-  g_variant_unref (platform_data);
+  app = g_initable_new (GTK_TYPE_APPLICATION, 
+			NULL,
+			&error,
+			"application-id", appid, 
+			"argv", argv_variant, 
+			NULL);
+  if (!app)
+    {
+      g_error ("%s", error->message);
+      g_clear_error (&error);
+      return NULL;
+    }
 
   return app;
 }
@@ -305,7 +369,12 @@ static gchar *default_title;
  * behaviour is to call gtk_application_quit().
  *
  * If your application uses only a single toplevel window, you can
- * use gtk_application_get_window().
+ * use gtk_application_get_window(). If you are using a sub-class
+ * of #GtkApplication you should call gtk_application_create_window()
+ * to let the #GtkApplication instance create a #GtkWindow and add
+ * it to the list of toplevels of the application. You should call
+ * this function only to add #GtkWindow<!-- -->s that you created
+ * directly using gtk_window_new().
  *
  * Since: 3.0
  */
@@ -313,13 +382,27 @@ void
 gtk_application_add_window (GtkApplication *app,
                             GtkWindow      *window)
 {
-  app->priv->windows = g_slist_prepend (app->priv->windows, window);
+  GtkApplicationPrivate *priv;
+
+  g_return_if_fail (GTK_IS_APPLICATION (app));
+  g_return_if_fail (GTK_IS_WINDOW (window));
+
+  priv = app->priv;
+
+  if (g_slist_find (priv->windows, window) != NULL)
+    return;
+
+  priv->windows = g_slist_prepend (priv->windows, window);
+
+  if (priv->default_window == NULL)
+    priv->default_window = window;
 
   if (gtk_window_get_title (window) == NULL && default_title != NULL)
     gtk_window_set_title (window, default_title);
 
   g_signal_connect (window, "destroy",
-                    G_CALLBACK (gtk_application_on_window_destroy), app);
+                    G_CALLBACK (gtk_application_on_window_destroy),
+                    app);
 }
 
 /**
@@ -338,7 +421,8 @@ gtk_application_add_window (GtkApplication *app,
  * If your application has more than one toplevel window (e.g. an
  * single-document-interface application with multiple open documents),
  * or if you are constructing your toplevel windows yourself (e.g. using
- * #GtkBuilder), use gtk_application_add_window() instead.
+ * #GtkBuilder), use gtk_application_create_window() or
+ * gtk_application_add_window() instead.
  *
  * Returns: (transfer none): The default #GtkWindow for this application
  *
@@ -347,15 +431,45 @@ gtk_application_add_window (GtkApplication *app,
 GtkWindow *
 gtk_application_get_window (GtkApplication *app)
 {
-  if (app->priv->default_window != NULL)
-    return app->priv->default_window;
+  GtkApplicationPrivate *priv;
+  GtkWindow *window;
 
-  app->priv->default_window = GTK_WINDOW (gtk_window_new (GTK_WINDOW_TOPLEVEL));
-  g_object_ref_sink (app->priv->default_window);
+  g_return_val_if_fail (GTK_IS_APPLICATION (app), NULL);
 
-  gtk_application_add_window (app, app->priv->default_window);
+  priv = app->priv;
 
-  return app->priv->default_window;
+  if (priv->default_window != NULL)
+    return priv->default_window;
+
+  return gtk_application_create_window (app);
+}
+
+/**
+ * gtk_application_create_window:
+ * @app: a #GtkApplication
+ *
+ * Creates a new #GtkWindow for the application.
+ *
+ * This function calls the #GtkApplication::create_window() virtual function,
+ * which can be overridden by sub-classes, for instance to use #GtkBuilder to
+ * create the user interface. After creating a new #GtkWindow instance, it will
+ * be added to the list of toplevels associated to the application.
+ *
+ * Return value: (transfer none): the newly created application #GtkWindow
+ *
+ * Since: 3.0
+ */
+GtkWindow *
+gtk_application_create_window (GtkApplication *app)
+{
+  GtkWindow *window;
+
+  g_return_val_if_fail (GTK_IS_APPLICATION (app), NULL);
+
+  window = GTK_APPLICATION_GET_CLASS (app)->create_window (app);
+  gtk_application_add_window (app, window);
+
+  return window;
 }
 
 /**
@@ -377,15 +491,32 @@ gtk_application_run (GtkApplication *app)
  * gtk_application_quit:
  * @app: a #GtkApplication
  *
- * Request the application exit.
- * By default, this method will exit the main loop; see gtk_main_quit().
+ * Request the application exit.  This function invokes
+ * g_application_quit_with_data(), which normally will
+ * in turn cause @app to emit #GtkApplication::quit.
+ *
+ * To control an application's quit behavior (for example, to ask for
+ * files to be saved), connect to the #GtkApplication::quit signal
+ * handler.
  *
  * Since: 3.0
  */
 void
 gtk_application_quit (GtkApplication *app)
 {
-  g_application_quit (G_APPLICATION (app), gtk_get_current_event_time ());
+  GVariantBuilder builder;
+  GVariant *platform_data;
+
+  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
+  g_variant_builder_add (&builder, "{sv}",
+			 "timestamp",
+			 g_variant_new ("u",
+					gtk_get_current_event_time ()));
+  platform_data = g_variant_builder_end (&builder);
+    
+  g_application_quit_with_data (G_APPLICATION (app), platform_data);
+
+  g_variant_unref (platform_data);
 }
 
 static void
@@ -460,6 +591,8 @@ gtk_application_init (GtkApplication *application)
 {
   application->priv = G_TYPE_INSTANCE_GET_PRIVATE (application, GTK_TYPE_APPLICATION, GtkApplicationPrivate);
 
+  g_object_set (application, "platform-data", gtk_application_format_activation_data (), NULL);
+
   setup_default_window_decorations ();
 }
 
@@ -495,10 +628,14 @@ gtk_application_class_init (GtkApplicationClass *klass)
   gobject_class->set_property = gtk_application_set_property;
 
   application_class->run = gtk_application_default_run;
-  application_class->quit = gtk_application_default_quit;
-  application_class->action = gtk_application_default_action;
+  application_class->quit_with_data = gtk_application_default_quit_with_data;
+  application_class->action_with_data = gtk_application_default_action_with_data;
   application_class->prepare_activation = gtk_application_default_prepare_activation;
 
+  klass->quit = gtk_application_default_quit;
+  klass->action = gtk_application_default_action;
+  klass->create_window = gtk_application_default_create_window;
+
   klass->activated = gtk_application_default_activated;
 
   /**
@@ -518,10 +655,57 @@ gtk_application_class_init (GtkApplicationClass *klass)
                   G_SIGNAL_RUN_LAST,
                   G_STRUCT_OFFSET (GtkApplicationClass, activated),
                   NULL, NULL,
-                  g_cclosure_marshal_VOID__BOXED,
+                  g_cclosure_marshal_VOID__VARIANT,
                   G_TYPE_NONE, 1,
                   G_TYPE_VARIANT);
 
+  /**
+   * GtkApplication::quit:
+   * @application: the object on which the signal is emitted
+   *
+   * This signal is emitted when a quit is initiated.  See also
+   * the #GApplication::quit-with-data signal which may in
+   * turn trigger this signal.
+   *
+   * The default handler for this signal exits the mainloop of the
+   * application.
+   *
+   * Returns: %TRUE if the signal has been handled, %FALSE to continue
+   *   signal emission
+   */
+  gtk_application_signals[QUIT] =
+    g_signal_new (g_intern_static_string ("quit"),
+                  G_OBJECT_CLASS_TYPE (klass),
+                  G_SIGNAL_RUN_LAST,
+                  G_STRUCT_OFFSET (GtkApplicationClass, quit),
+                  g_signal_accumulator_true_handled, NULL,
+                  _gtk_marshal_BOOLEAN__VOID,
+                  G_TYPE_BOOLEAN, 0);
+
+  /**
+   * GtkApplication::action:
+   * @application: the object on which the signal is emitted
+   * @name: The name of the activated action
+   *
+   * This signal is emitted when an action is activated. The action name
+   * is passed as the first argument, but also as signal detail, so it
+   * is possible to connect to this signal for individual actions.
+   *
+   * See also the #GApplication::action-with-data signal which may in
+   * turn trigger this signal.
+   *
+   * The signal is never emitted for disabled actions.
+   */
+  gtk_application_signals[ACTION] =
+    g_signal_new (g_intern_static_string ("action"),
+                  G_OBJECT_CLASS_TYPE (klass),
+                  G_SIGNAL_RUN_FIRST | G_SIGNAL_NO_RECURSE | G_SIGNAL_DETAILED,
+                  G_STRUCT_OFFSET (GtkApplicationClass, action),
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__STRING,
+                  G_TYPE_NONE, 1,
+                  G_TYPE_STRING);
+                 
   g_type_class_add_private (gobject_class, sizeof (GtkApplicationPrivate));
 }
 
